################################################################
#
#   Assembly Line - Machine
#
################################################################

from random import randint
from pygame import Rect
from transformation import Transformation, add, rotate_vector

class MachineNotConfiguredError(Exception):
	pass

#----------------------------------------------------------------------

class MachineType(type):
	#  title                 string
	#  cost                  int
	#  months_to_develop     int
	#  local_bounding_rect   Rect
	#  local_covering_rects  list[layer] of [Rect]
	
	months_to_develop = 2

	local_feedback_rects = []
	
	def bounding_rect(self, t):
		lr = self.local_bounding_rect
		gr = t.rect(lr)
		#print "MachineType.global_bounding_rect:" ###
		#print "rotation =", t.rotation, "translation =", t.translation ###
		#print "lr =", lr, "gr =", gr ###
		return gr
	
	def covered_rects(self, t):
		return [
			[t.rect(r) for r in rects]
				for rects in self.local_covered_rects]
	
	def feedback_rects(self, t):
		return [t.rect(r) for r in self.local_feedback_rects]

#----------------------------------------------------------------------

class Machine(object):
	#  translation           vector
	#  rotation              int
	#  purchased             boolean
	#  state                 string
	#  timeout               int
	#
	#  Class attributes:
	#
	#  rotation_mask         int
	#  on_left               boolean
	#  layer                 int
	#  cost                  int
	#  is_conveyor           boolean
	#  is_conveyor_end       boolean
	#  is_conveyor_drive     boolean
	#  is_inspectable        boolean
	#  editor_class          Widget class
	
	__metaclass__ = MachineType

	position_tolerance = 4

	rotatable = False
	on_left = False
	is_conveyor = False
	is_conveyor_end = False
	is_conveyor_drive = False
	is_inspectable = False
	editor_class = None
	rotation_mask = 0
	resettable = False
	
	translation = (0, 0)
	rotation = 0
	purchased = False
	state = 'idle'
	timeout = 0
	
	level_gauge_bg_color = (0, 0, 0)
	level_gauge_fg_color = (0, 255, 0)
	level_gauge_border_color = (128, 128, 128)
	level_gauge_border_width = 2
	
	def local_to_global_transformation(self):
		return Transformation(self.rotation, self.translation)

	def global_bounding_rect(self):
		t = self.local_to_global_transformation()
		return self.__class__.bounding_rect(t)
	
	def machine_bounding_rect(self, t):
		return self.__class__.bounding_rect(t.trans(self))

	def global_covered_rects(self):
		t = self.local_to_global_transformation()
		return self.__class__.covered_rects(t)
	
	def global_vector(self, v):
		return rotate_vector(self.rotation, v)
	
	def global_origin(self):
		return self.translation
	
	def global_point(self, p):
		t = self.local_to_global_transformation()
		return t.point(p)

	def global_rect(self, r):
		t = self.local_to_global_transformation()
		return t.rect(r)
	
	def position_global(self, p):
		self.translation = p
	
	def set_transformation(self, t):
		self.rotation = t.rotation
		self.translation = t.translation
	
	def offset_global(self, d):
		ld = rotate_vector((-self.rotation) % 4, d)
		self.translation = add(ld, self.translation)

	def local_topleft(self):
		return self.local_bounding_rect.topleft
	
	def in_layer(self, layer):
		return self.layer == layer
	
	def draw(self, surf, t, layer):
		self.draw_local_layer(surf, t.trans(self), layer)
	
	def draw_local_layer(self, surf, t, layer):
		self.draw_local(surf, t)

	def begin_frame(self, factory):
		pass

	def check_ready(self):
		pass

	def handle_part(self, factory, part):
		pass

	def update_neighbourhood(self, factory):
		pass

	def part_removed(self, part):
		pass

	def removed_from_factory(self, factory):
		self.reset(factory)
	
	def reset(self, factory):
		self.state = 'idle'
		self.timeout = 0

	def flush(self, factory, attr):
		part = getattr(self, attr)
		if part:
			part.handler = None
			setattr(self, attr, None)
			self.position_flushed_part(factory, part)
	
	def position_flushed_part(self, factory, part):
		pass

	def find_unpurchased_components(self, result):
		pass
	
	def extract_components(self, result):
		pass
	
	def extract_crates(self, crates):
		pass

	def part_in_rect(self, part, local_rect):
		r = self.global_rect(local_rect)
		return r.collidepoint(part.position)

	def move_part_to(self, factory, part, local_point):
		p = self.global_point(local_point)
		factory.move_part_to(part, p)
	
	def scatter_part(self, factory, part, (x, y), spread):
		def r():
			return randint(-spread, spread) + randint(-spread, spread)
		dx = r()
		dy = r()
		self.move_part_to(factory, part, (x + dx, y + dy))
		part.handler = None

	def part_in_input_rect(self, part):
		p = part.position
		r = self.global_rect(self.local_input_rect)
		return r.collidepoint(p)

	def begin_frame(self, factory):
		if self.timeout > 0:
			self.timeout -= 1
		else:
			self.change_state(factory)
	
	def change_state(self, factory):
		pass

	def draw_level_gauge(self, surf, t, cur_level, max_level):
		#print "Machine.draw_level_gauge:", cur_level, "of", max_level ###
		local_frame = self.level_gauge_rect
		r = t.rect(local_frame)
		surf.fill(self.level_gauge_border_color, r)
		b = self.level_gauge_border_width
		d = - 2 * b
		dd = (d, d)
		r.inflate_ip(dd)
		surf.fill(self.level_gauge_bg_color, r)
		local_level = local_frame.inflate(dd)
		h = cur_level * local_level.height // max_level
		local_level.top = local_level.bottom - h
		local_level.height = h
		r = t.rect(local_level)
		surf.fill(self.level_gauge_fg_color, r)
